"""Add is_task to workflow.

Revision ID: 161918133bec
Revises: 68d14db1b8da
Create Date: 2025-05-08 11:25:51.966410

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "161918133bec"
down_revision = "68d14db1b8da"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workflows", sa.Column("is_task", sa.Boolean(), server_default=sa.text("false"), nullable=False))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workflows", "is_task")
    # ### end Alembic commands ###
